#include "Pokitto.h"
#include "Grain.h"
#include <array>
#include <stdio.h>
#include "mbed.h"

#define W 55
#define H 44
#define DUMP 0.35
#define GRAINS 128
#define MAX_SERIAL_BUFFER 20
#define AVARAGES 10

using PC = Pokitto::Core;
using PD = Pokitto::Display;
using PB = Pokitto::Buttons;

Point force;
float angle;
float g;
std::array < Grain, GRAINS > sands;

Serial ser(EXT7, EXT6);
static char serialInBuffer[MAX_SERIAL_BUFFER];
static int serialCount = 0;
int accelerometer[3][AVARAGES];
bool testMode = false;

void Rx_interrupt() {
    while (ser.readable()) {
        char byteIn = ser.getc();
        serialInBuffer[serialCount] = byteIn; // store the character
        if (serialCount < MAX_SERIAL_BUFFER) // increase the counter.
            serialCount++;
        if ((byteIn == '\n')) { // if an end of line is found
            serialInBuffer[serialCount] == '\0'; // null terminate the input
            int x, y, z;
            if (sscanf(serialInBuffer, "x%dy%dz%d#\r\n", & x, & y, & z) == 3) { // managed to read all 3 values
                //Shift array to the right
                for (int i = AVARAGES - 2; i >= 0; i--) {
                    accelerometer[0][i + 1] = accelerometer[0][i];
                    accelerometer[1][i + 1] = accelerometer[1][i];
                    accelerometer[2][i + 1] = accelerometer[2][i];
                }
                accelerometer[0][0] = x;
                accelerometer[1][0] = y;
                accelerometer[2][0] = z;
            }
            serialCount = 0; // reset the buffer
        }
    }
}


void init() {
    ser.baud(115200);
    ser.attach( & Rx_interrupt, Serial::RxIrq);

    angle = 0;
    g = 0.05;
    for (int i = 0; i < sands.size(); i++) {

        bool placeFree = true;
        do {
            placeFree = true;
            sands[i].pos.x = random(2, W - 2);
            sands[i].pos.y = random(2, H - 2);
            for (int j = 0; j < sands.size(); j++) {
                if (i == j) continue;
                if (sands[j].pos.x.getInteger() == sands[i].pos.x.getInteger() &&
                    sands[j].pos.y.getInteger() == sands[i].pos.y.getInteger()) {
                    placeFree = false;
                    break;
                }
            }
        } while (!placeFree);
        //----
        sands[i].vel.x = 0;
        sands[i].vel.y = 0;
    }
}


void update() {

    //Read inputs
    if (PB::leftBtn()) {
        angle -= 0.1;
    }
    if (PB::rightBtn()) {
        angle += 0.1;
    }
    if (PB::pressed(BTN_B)) {
        init();
    }
    if (PB::pressed(BTN_C)) {
        testMode = !testMode;
    }

    if (testMode) {
        force.x = g * sin(angle);
        force.y = g * cos(angle);
    } else {
        float xSum = 0;
        float ySum = 0;
        float zSum = 0;
        for (int i = 0; i < AVARAGES; i++) {
            xSum += accelerometer[0][i];
            ySum += accelerometer[1][i];
            zSum += accelerometer[2][i];
        }
        force.x = (float)(xSum / AVARAGES) / 10000.0;
        force.y = (float)(ySum / AVARAGES) / 10000.0;
    }


    for (int i = 0; i < sands.size(); i++) {
        auto & grain = sands[i];

        //Save position
        auto oldXf = grain.pos.x;
        auto oldYf = grain.pos.y;
        auto oldXi = oldXf.getInteger();
        auto oldYi = oldYf.getInteger();

        //Apply force
        grain.vel.x += force.x + (random(-10, 10) / 1000.0);
        grain.vel.y += force.y + (random(-10, 10) / 1000.0);

        //Limit force
        if (grain.vel.x > 1) grain.vel.x = 1;
        if (grain.vel.x < -1) grain.vel.x = -1;
        if (grain.vel.y > 1) grain.vel.y = 1;
        if (grain.vel.y < -1) grain.vel.y = -1;

        //Calc new position
        auto newXf = grain.pos.x + grain.vel.x;
        auto newYf = grain.pos.y + grain.vel.y;

        //------------------------------------
        //Border limits
        if (newXf > W - 2) {
            newXf = W - 2;
            grain.vel.x = -grain.vel.x * DUMP;
        }
        if (newXf < 2) {
            newXf = 2;
            grain.vel.x = -grain.vel.x * DUMP;
        }
        if (newYf > H - 2) {
            newYf = H - 2;
            grain.vel.y = -grain.vel.y * DUMP;
        }
        if (newYf < 2) {
            newYf = 2;
            grain.vel.y = -grain.vel.y * DUMP;
        }
        //------------------------------------


        auto newXi = newXf.getInteger();
        auto newYi = newYf.getInteger();

        bool collideXY = false;
        bool collideX = false;
        bool collideY = false;
        for (int j = 0; j < sands.size(); j++) {
            if (i == j) continue;
            auto & oGrain = sands[j];
            if (newXi == oGrain.pos.x.getInteger() && newYi == oGrain.pos.y.getInteger()) {
                collideXY = true;
                //Chek if moving only x would be free
                for (int jx = 0; jx < sands.size(); jx++) {
                    if (i == jx) continue;
                    auto & xGrain = sands[jx];
                    if (newXi == xGrain.pos.x.getInteger() && oldYi == xGrain.pos.y.getInteger()) {
                        collideX = true;
                        break;
                    }
                }
                //Chek if moving only y would be free
                for (int jy = 0; jy < sands.size(); jy++) {
                    if (i == jy) continue;
                    auto & yGrain = sands[jy];
                    if (oldXi == yGrain.pos.x.getInteger() && newYi == yGrain.pos.y.getInteger()) {
                        collideY = true;
                        break;
                    }
                }
                break;
            }
        }

        //Resolve collision
        if (collideXY && !collideX) {
            newYf = oldYf;
            grain.vel.y = -grain.vel.y * DUMP;
        } else if (collideXY && !collideY) {
            newXf = oldXf;
            grain.vel.x = -grain.vel.x * DUMP;
        } else if (collideXY && collideX && collideY) {
            newXf = oldXf;
            newYf = oldYf;
            grain.vel.x = -grain.vel.x * DUMP;
            grain.vel.y = -grain.vel.y * DUMP;
        }

        //Save grain pos
        grain.pos.x = newXf;
        grain.pos.y = newYf;

        //Draw
        PD::setColor(i / 64 ? 3 : 7);
        PD::fillRect(newXf.getInteger() * 2, newYf.getInteger() * 2, 2, 1);
    }


    //--------------- Draw force vector
    int xl = (int)((float) force.x * 1200.0);
    int yl = (int)((float) force.y * 1200.0);
    PD::setColor(1);
    PD::drawLine(W, H, W + xl, H + yl);
}